`timescale 1ns/1ps
//Address 0  (0x00) returns the status of the LEDS in bits 3 downto 0
//Address 1  (0x01) returns the status of the Buttons in bits 4 downto 0

//Stores to anything other than 0x00 do nothing

module MemoryMappedBoardIO(
	input CLK,
	input Reset,
	input [5:0] IOAddress,
	input [31:0] IOData,
	input IOWE,
	input [4:0] nButtonInputs,
	output wire [31:0] IODataOut,
	output wire [3:0] nLEDBus) ;
	
	reg [4:0] ButtonReg;
	reg [3:0] LEDS;
	wire LED_ADDR = (IOAddress == 6'b000000);
	wire LED_LD = (LED_ADDR & IOWE);
	
	assign nLEDBus[3:0] = ~(LEDS[3:0]);
	assign IODataOut[31:16] = 16'b0;
	assign IODataOut[15:8] = {3'b0, ButtonReg[4:0]};
	assign IODataOut[7:0] = {4'b0, LEDS[3:0]}; //lowest byte gets LED status
	
		
	always @(posedge CLK) begin
			LEDS <= (Reset)? 4'h0 : ((LED_LD)? IOData[3:0] : LEDS);
			ButtonReg <= (Reset)? 5'b00000 : ~(nButtonInputs[4:0]);
	end
	
endmodule