`ifdef _VCP
`else
`define library(a,b)
`endif

`timescale 1ps / 1ps

module Memory_Toplevel (
	CLK,
	DataIn,
	IODev1Input,
	IODev2Input,
	IODev3Input,
	IODev4Input,
	WriteEnable,
	Size,
	SignedLoad,
	DataAddress,
	InstrAddress,
	DataOut,
	IODev1WE,
	IODev2WE,
	IODev3WE,
	IODev4WE,
	CharWE,
	ColorWE,
	Instruction
);
// ------------ Port declarations --------- //
input CLK;
wire CLK;
input wire [31:0] DataIn,IODev1Input,IODev2Input,IODev3Input,IODev4Input,DataAddress,InstrAddress ;
input wire WriteEnable, SignedLoad;
input wire [1:0] Size;
output wire [31:0] DataOut,Instruction;
output wire IODev1WE, IODev2WE, IODev3WE,IODev4WE;
output wire CharWE,ColorWE;

parameter BIOS_INIT_FILE = "$DSN\\src\\bios_init.hex";
parameter STACK_INIT_FILE = "$DSN\\src\\stack_init.hex";
parameter DATA_INIT_FILE = "$DSN\\src\\data_init.hex"; 

wire InstSrc;
supply1 VCC;
wire [31:0] BIOSOut;
wire [9:0] dAddress;
wire [1:0] dataTarget;
wire DataWE,IOWE,StackWE;
wire [31:0] DRAMOut, DataOutRaw;
wire [11:0] iAddress;
wire InstWE,InstWE_BOT,InstWE_LO,InstWE_MED,InstWE_HI;
wire [7:0] IOAddress;
wire [31:0] IOOut;
wire [31:0] IRAMOut,IRAMOut_HI,IRAMOut_MED,IRAMOut_LO,IRAMOut_BOT;
wire [7:0] ROMAddress;
wire [9:0] StackAddress;
wire [31:0] StackOut;
wire BadAddress;
wire [31:0] ErrorAddress;

assign ErrorAddress = 32'h08000029;

assign IRAMOut[31:0] = ((iAddress[11])? ((iAddress[10])? IRAMOut_HI[31:0] : IRAMOut_MED[31:0]) :
										((iAddress[10])? IRAMOut_LO[31:0] : IRAMOut_BOT));

dmVw_1_4_bot #(.width(1)) InstRamControl
(
	.in(InstWE),
	.out0(InstWE_BOT),
	.out1(InstWE_LO),
	.out2(InstWE_MED),
	.out3(InstWE_HI),
	.sel(iAddress[11:10])
);

BIOSROM256x32 #(.INIT_FILE(BIOS_INIT_FILE)) BIOS 
(
	.ADDR(ROMAddress),
	.OUTPUT(BIOSOut)
);

Ram256x32_sw_ar_wb #(.INIT_FILE(DATA_INIT_FILE)) DRAM
(
	.ADDR(dAddress),
	.CLK(CLK),
	.DI(DataIn),
	.DO(DRAMOut),
	.SIZE(Size),
	.WE(DataWE)
);

dmVw_1_4_bot #(.width(1)) IOControl
(
	.in(IOWE),
	.out0(IODev1WE),
	.out1(IODev2WE),
	.out2(IODev3WE),
	.out3(IODev4WE),
	.sel(IOAddress[7:6])
);

m32w_4_1_bot IOInputMux
(
	.in0(IODev1Input[31:0]),
	.in1(IODev2Input[31:0]),
	.in2(IODev3Input[31:0]),
	.in3(IODev4Input[31:0]),
	.out(IOOut[31:0]),
	.sel(IOAddress[7:6])
);

m32w_4_1_bot InstMux
(
	.in0(IRAMOut[31:0]),
	.in1(BIOSOut[31:0]),
	.in2(ErrorAddress[31:0]),
	.in3(ErrorAddress[31:0]),
	.out(Instruction[31:0]),
	.sel({1'b0,InstSrc})
);

MMU MMU
(
	.DataAddress(dAddress),
	.DataWE(DataWE),
	.IOAddress(IOAddress),
	.IOWE(IOWE),
	.IRAMAddress(iAddress),
	.InstWE(InstWE),
	.ROMAddress(ROMAddress),
	.StackAddress(StackAddress),
	.StackWE(StackWE),
	.dataAddress(DataAddress),
	.dataTarget(dataTarget),
	.instFromROM(InstSrc),
	.PC(InstrAddress),
	.WE(WriteEnable),
	.Size(Size),
	.CharWE(CharWE),
	.ColorWE(ColorWE),
	.BadAddress(BadAddress)
);



m32w_4_1_bot OutputsourceMux
(
	.in0(IRAMOut[31:0]),
	.in1(DRAMOut[31:0]),
	.in2(StackOut[31:0]),
	.in3(IOOut[31:0]),
	.out(DataOutRaw[31:0]),
	.sel(dataTarget)
);

LoadUnit LoadAdapter (.DataIn(DataOutRaw[31:0]),
	.DataOut(DataOut[31:0]),
	.size(Size[1:0]),
	.offset(DataAddress[1:0]),
	.isSigned(SignedLoad)
);

Ram256x32_sw_ar_wb #(.INIT_FILE(STACK_INIT_FILE)) Stack
(
	.ADDR(StackAddress),
	.CLK(CLK),
	.DI(DataIn),
	.DO(StackOut),
	.SIZE(Size),
	.WE(StackWE)
);



Ram256x32_sw_ar_wb IRAM_HI
(
	.ADDR(iAddress[9:0]),
	.CLK(CLK),
	.DI(DataIn),
	.DO(IRAMOut_HI),
	.SIZE(Size),
	.WE(InstWE_HI)
);

Ram256x32_sw_ar_wb IRAM_MED
(
	.ADDR(iAddress[9:0]),
	.CLK(CLK),
	.DI(DataIn),
	.DO(IRAMOut_MED),
	.SIZE(Size),
	.WE(InstWE_MED)
);

Ram256x32_sw_ar_wb IRAM_LO
(
	.ADDR(iAddress[9:0]),
	.CLK(CLK),
	.DI(DataIn),
	.DO(IRAMOut_LO),
	.SIZE(Size),
	.WE(InstWE_LO)
);

Ram256x32_sw_ar_wb IRAM_BOT
(
	.ADDR(iAddress[9:0]),
	.CLK(CLK),
	.DI(DataIn),
	.DO(IRAMOut_BOT),
	.SIZE(Size),
	.WE(InstWE_BOT)
);

endmodule 

module LoadUnit(
	DataIn,
	DataOut,
	size,
	offset,
	isSigned
) ;																						  

input wire [31:0] DataIn;
input wire [1:0] offset,size;
input wire isSigned;
output wire [31:0] DataOut;

wire byte_len = ~size[1] & size[0];
wire half_len = size[1] & ~size[0];
wire word_len = size[1] & size[0];
wire neg_byte = (DataIn[7] & ~offset[1] & ~offset[0])|(DataIn[15] & ~offset[1] & offset[0])|(DataIn[23] & offset[1] & ~offset[0])|(DataIn[31] & offset[1] & offset[0]);
wire neg_half = (DataIn[15] & ~offset[1]) | (DataIn[31] & offset[1]);

wire [15:0] half_data = (offset[1])? DataIn[31:16] : DataIn[15:0];
wire [7:0] byte_data = (offset[1])? (offset[0])? DataIn[31:24] : DataIn[23:16] : (offset[0])? DataIn[15:8] : DataIn[7:0];


assign DataOut[31:16] = (word_len)? DataIn[31:16] : {16{((half_len & neg_half)|(byte_len & neg_byte)) & isSigned}};
assign DataOut[15:8] = (word_len)? DataIn[15:8] : (half_len)? half_data[15:8] : {8{(byte_len & neg_byte & isSigned)}};
assign DataOut[7:0] = (word_len)? DataIn[7:0] : (half_len)? half_data[7:0] : byte_data;

endmodule	
	
module MMU(PC,
	dataAddress,
	WE,
	ROMAddress,
	IRAMAddress,
	DataAddress,
	StackAddress,
	IOAddress,
	Size,
	InstWE,
	DataWE,
	StackWE,
	IOWE,
	CharWE,
	ColorWE,
	dataTarget,
	instFromROM,
	BadAddress
) ;
	
	input wire [31:0] PC;
	input wire [31:0] dataAddress;
	input wire WE;
	input wire [1:0] Size;
	
	output wire [7:0] ROMAddress;
	output wire [11:0] IRAMAddress;
	output wire [9:0] DataAddress;
	output wire [9:0] StackAddress;
	output wire [7:0] IOAddress;
	
	output wire InstWE;
	output wire DataWE;
	output wire StackWE;
	output wire IOWE;
	output wire [1:0] dataTarget;
	output wire CharWE,ColorWE;
	output wire instFromROM,BadAddress;		
							 
	parameter BIOS_HIGH  = 22'b0000_0000_0000_0000_0000_00;
	parameter IRAM_HIGH  = 20'h00400;
	parameter DRAM_HIGH  = 22'b0001_0000_0000_0000_0000_00;
	parameter STACK_HIGH = 22'b0111_1111_1111_1111_1111_11;
	parameter IO_HIGH    = 24'h8000_00;
	parameter VGA_CHAR_HIGH  = 19'b0000_0000_0000_0000_110;	// VGA CHAR Addresses 0x0000C000-0x0000DFFF
	parameter VGA_COLOR_HIGH = 19'b0000_0000_0000_0000_111;	 // VGA Color Addresses 0x0000E000-0x0000FFFF
	
	wire BIOS_RANGE = (dataAddress[31:10] == BIOS_HIGH);
	wire IRAM_RANGE = (dataAddress[31:12] == IRAM_HIGH);
	wire DRAM_RANGE = (dataAddress[31:10] == DRAM_HIGH);
	wire STACK_RANGE = (dataAddress[31:10] == STACK_HIGH); 
	wire IO_RANGE = (dataAddress[31:8] == IO_HIGH);
	wire VGA_CHAR_RANGE = (dataAddress[31:13] == VGA_CHAR_HIGH);
	wire VGA_COLOR_RANGE = (dataAddress[31:13] == VGA_COLOR_HIGH);
	
	assign dataTarget = {(STACK_RANGE)|(IO_RANGE),(DRAM_RANGE)|(IO_RANGE)};
	assign BadAddress = ~(BIOS_RANGE|IRAM_RANGE|DRAM_RANGE|STACK_RANGE|IO_RANGE|VGA_CHAR_RANGE|VGA_COLOR_RANGE);
	assign instFromROM = (PC[31:10] == BIOS_HIGH);
	assign ROMAddress = PC[9:2];
	assign IRAMAddress = (instFromROM)? dataAddress[11:0] : PC[11:0];
	assign DataAddress = dataAddress[9:0];
	assign StackAddress = dataAddress[9:0];
	assign IOAddress = dataAddress[7:0];
	
	assign InstWE = ((IRAM_RANGE) & (instFromROM) & WE );
	assign DataWE = (DRAM_RANGE) & WE;
	assign StackWE = (STACK_RANGE) & WE;
	assign IOWE = (IO_RANGE) & WE;
	assign CharWE = (VGA_CHAR_RANGE) & (Size[0] & ~Size[1]) & WE;
	assign ColorWE = (VGA_COLOR_RANGE) & (Size[0] & ~Size[1]) & WE;

endmodule

module BIOSROM256x32 (
	input [7:0] ADDR,
	output [31:0] OUTPUT
	) ;

parameter INIT_FILE="$DSN\\src\\bios.hex";

`ifdef SYNTHESIS           


            `library("U1_ARRAY0","ovi_unisim")
            `library("U1_ARRAY1","ovi_unisim")
            `library("U1_ARRAY2","ovi_unisim")
            `library("U1_ARRAY3","ovi_unisim")
            `library("U1_ARRAY4","ovi_unisim")
            `library("U1_ARRAY5","ovi_unisim")
            `library("U1_ARRAY6","ovi_unisim")
            `library("U1_ARRAY7","ovi_unisim")
            `library("U1_ARRAY8","ovi_unisim")
            `library("U1_ARRAY9","ovi_unisim")
            `library("U1_ARRAY10","ovi_unisim")
            `library("U1_ARRAY11","ovi_unisim")
            `library("U1_ARRAY12","ovi_unisim")
            `library("U1_ARRAY13","ovi_unisim")
            `library("U1_ARRAY14","ovi_unisim")
            `library("U1_ARRAY15","ovi_unisim")
            `library("U1_ARRAY16","ovi_unisim")
            `library("U1_ARRAY17","ovi_unisim")
            `library("U1_ARRAY18","ovi_unisim")
            `library("U1_ARRAY19","ovi_unisim")
            `library("U1_ARRAY20","ovi_unisim")
            `library("U1_ARRAY21","ovi_unisim")
            `library("U1_ARRAY22","ovi_unisim")
            `library("U1_ARRAY23","ovi_unisim")
            `library("U1_ARRAY24","ovi_unisim")
            `library("U1_ARRAY25","ovi_unisim")
            `library("U1_ARRAY26","ovi_unisim")
            `library("U1_ARRAY27","ovi_unisim")
            `library("U1_ARRAY28","ovi_unisim")
            `library("U1_ARRAY29","ovi_unisim")
			`library("U1_ARRAY30","ovi_unisim")
            `library("U1_ARRAY31","ovi_unisim")
         
            ROM256X1 U1_ARRAY31	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[31]}));
            ROM256X1 U1_ARRAY30	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[30]}));
            ROM256X1 U1_ARRAY29	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[29]}));
            ROM256X1 U1_ARRAY28	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[28]}));
            ROM256X1 U1_ARRAY27	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[27]}));
            ROM256X1 U1_ARRAY26	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[26]}));
            ROM256X1 U1_ARRAY25	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[25]}));
            ROM256X1 U1_ARRAY24	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[24]}));
            ROM256X1 U1_ARRAY23	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[23]}));
            ROM256X1 U1_ARRAY22 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[22]}));
            ROM256X1 U1_ARRAY21	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[21]}));
            ROM256X1 U1_ARRAY20	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[20]}));
            ROM256X1 U1_ARRAY19 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[19]}));
            ROM256X1 U1_ARRAY18 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[18]}));
            ROM256X1 U1_ARRAY17	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[17]}));
            ROM256X1 U1_ARRAY16	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[16]}));
            ROM256X1 U1_ARRAY15	(.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[15]}));
            ROM256X1 U1_ARRAY14 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[14]}));
            ROM256X1 U1_ARRAY13 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[13]}));
            ROM256X1 U1_ARRAY12 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[12]}));
            ROM256X1 U1_ARRAY11 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[11]}));
            ROM256X1 U1_ARRAY10 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[10]}));
            ROM256X1 U1_ARRAY9 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[9]}))  ;
            ROM256X1 U1_ARRAY8 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[8]}))  ;
            ROM256X1 U1_ARRAY7 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[7]}))  ;
            ROM256X1 U1_ARRAY6 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[6]}))  ;
            ROM256X1 U1_ARRAY5 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[5]}))  ;
            ROM256X1 U1_ARRAY4 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[4]}))  ;
            ROM256X1 U1_ARRAY3 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[3]}))  ;
            ROM256X1 U1_ARRAY2 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[2]}))  ;
            ROM256X1 U1_ARRAY1 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[1]}))  ;
            ROM256X1 U1_ARRAY0 (.A0({ADDR[0]}),.A1({ADDR[1]}),.A2({ADDR[2]}),.A3({ADDR[3]}),.A4({ADDR[4]}),.A5({ADDR[5]}),.A6({ADDR[6]}),.A7({ADDR[7]}),.O({OUTPUT[0]}))  ; 
`else 
	reg [31:0] ROM [0:255];
	
	initial begin
		$readmemh(INIT_FILE,ROM);
	end
		
	assign 	OUTPUT = ROM[ADDR];
`endif 

endmodule 

/*Top Level Memory System*/
module CacheMemorySytem_1Word(
	input CLK, Reset, 
	input DCacheReadRequest, 
	input DCacheWriteRequest, 
	input [29:0] DCacheAddr,
	input [31:0] DCacheDataIn, 
	output [31:0] DCacheDataOut,
	output DCacheValid,
	output DCacheInvalidAddress,
	
	input ICacheReadRequest, 
	input [29:0] ICacheAddr,
	output [31:0] ICacheDataOut,
	output ICacheValid,
	output ICacheInvalidAddress,
	
	
	input BypassReadRequest,
	input [1:0] BypassWrite,
	input [31:0] BypassAddr,
	output [31:0] BypassDataOut,
	output BypassInvalidAddress,
	
	
	output [1:0] IODev0_WE, IODev1_WE, IODev2_WE, IODev3_WE, IOVGAChar_WE, IOVGAColor_WE,
	input [31:0] IODev0_DataIn, IODev1_DataIn, IODev2_DataIn, IODev3_DataIn
	) /* synthesis syn_black_box */ ;
// synopsys translate_off	
	wire CacheInvalidAddress;
	assign DCacheInvalidAddress = CacheInvalidAddress;
	assign ICacheInvalidAddress = CacheInvalidAddress; 
	
	wire IO0_SEL,IO1_SEL,IO2_SEL,IO3_SEL,IOVGAChar_SEL,IOVGAColor_SEL;
	
	IOMemory io (
	.Addr(BypassAddr),
	.Read(BypassReadRequest),
	.Write(BypassWriteRequest),
	.InvalidAddress(BypassInvalidAddress),
	.IODev0_SEL(IO0_SEL),
	.IODev1_SEL(IO1_SEL),
	.IODev2_SEL(IO2_SEL),
	.IODev3_SEL(IO3_SEL),
	.IOVGAChar_SEL(IOVGAChar_SEL),
	.IOVGAColor_SEL(IOVGAColor_SEL)
	);
	
	CacheMemoryInterface_1Word cache (
	.CLK(CLK),
	.Reset(Reset),
	.DReadRequest(DCacheReadRequest),
	.IReadRequest(ICacheReadRequest),
	.DWriteRequest(DCacheWriteRequest),
	.DAddr(DCacheAddr),
	.IAddr(ICacheAddr),			 
	.DCacheLineIn(DCacheDataIn),
	.DCacheLineOut(DCacheDataOut),
	.ICacheLineOut(ICacheDataOut),
	.DValid(DCacheValid),
	.IValid(ICacheValid),
	.InvalidAddress(CacheInvalidAddress)
	); 
	
	
	//logic below here to produce proper IO signals
	//--------------------------------------------------
	
	assign IODev0_WE = {IO0_SEL && BypassWrite[1], IO0_SEL && BypassWrite[0]}; 
	assign IODev1_WE = {IO1_SEL && BypassWrite[1], IO1_SEL && BypassWrite[0]}; 
	assign IODev2_WE = {IO2_SEL && BypassWrite[1], IO2_SEL && BypassWrite[0]}; 
	assign IODev3_WE = {IO3_SEL && BypassWrite[1], IO3_SEL && BypassWrite[0]}; 
	
	assign IOVGAChar_WE = {IOVGAChar_SEL && BypassWrite[1], IOVGAChar_SEL && BypassWrite[0]};
	assign IOVGAColor_WE = {IOVGAColor_SEL && BypassWrite[1], IOVGAColor_SEL && BypassWrite[0]};
	
	assign BypassDataOut = IO0_SEL ? IODev0_DataIn : (IO1_SEL ? IODev1_DataIn : (IO2_SEL ? IODev2_DataIn : IODev3_DataIn));
// synopsys translate_on	
endmodule

/*Assert Read/Write Request high for one clock cycle and then read result when valid line goes high.*/
module CacheMemoryInterface_1Word(input CLK, Reset, 
	input DReadRequest, DWriteRequest,
	input [29:0] DAddr, input [31:0] DCacheLineIn, output [31:0] DCacheLineOut, output DValid, output InvalidAddress,
	input IReadRequest,
	input [29:0] IAddr, output [31:0] ICacheLineOut, output IValid
) /* synthesis syn_black_box */ ;
	
// synopsys translate_off	

	localparam NOREQUEST=1,
	DELAY=2,
	VALID=4;
	
	reg [4:0] Istate;
	reg [4:0] Dstate;
	
	reg IValid, DValid;
	
	reg WriteEnable;
	reg [29:0] Address;			 
	reg [29:0] IAddress;
	reg [29:0] DAddress;
	
	reg [31:0] WriteData;
	reg [31:0] DWriteData;
	reg	DWriteEnable; 
	
	wire [31:0] ReadData;
	
	wire DCacheMemoryInvalidAddress;
	
	assign DCacheLineOut = ReadData;
	assign ICacheLineOut = ReadData;
	
	
	always @(posedge CLK or posedge Reset) begin
		if (Reset) begin
			Istate<=NOREQUEST;
			Dstate<=NOREQUEST;
			IAddress<=30'b0;
			DAddress<=30'b0;
			IValid<=1'b0;
			DValid<=1'b0;
			WriteEnable<= 1'b0;
			WriteData<=32'b0;
			DWriteData<=32'b0;
			DWriteEnable<=1'b0;
			Address<=30'b0;
		end	else begin
		
		case (Istate)
			NOREQUEST: begin
				IValid<=1'b0;
				if (Dstate==DELAY && IReadRequest) begin
					//there's a data cache request that was delayed for an instruction cache request, so let that finish first
					Istate<=DELAY;
					IAddress<=IAddr;
				end else if (Dstate!=DELAY && IReadRequest) begin
					//no data request in progress, go ahead and request directly
					Istate<=VALID;
					Address<=IAddr;
					WriteEnable<=1'b0;
				end
			end
			DELAY: begin			//we wanted to wait a cycle before requesting the data, so go ahead and request now
				Istate<=VALID;
				Address<=IAddress;
				WriteEnable<=1'b0;
				IValid<=1'b0;
			end				  	  
			VALID: begin			//read request completed
				Istate<=NOREQUEST;
				IValid<=1'b1;
			end
		endcase
		
		case (Dstate)
			NOREQUEST: begin	
				DValid<=1'b0;
				if (Istate==DELAY && (DReadRequest || DWriteRequest)) begin
					//there a delayed instruction request, finish that first, delay the data request
					Dstate<=DELAY;
					DAddress<=DAddr;
					DWriteData<=DCacheLineIn;
					DWriteEnable<=DWriteRequest;
				end else if (Istate!=DELAY && (DReadRequest || DWriteRequest) && IReadRequest) begin 
					//no instructions queued, but we give priority to the instruction cache, so delay the instruction request
					Dstate<=DELAY;
					DAddress<=DAddr;
					DWriteData<=DCacheLineIn;
					DWriteEnable<=DWriteRequest;
				end else if (Istate!=DELAY && (DReadRequest || DWriteRequest) && ~IReadRequest) begin 
					//data request can execute immediately
					Dstate<=VALID;
					Address<=DAddr;
					WriteData<=DCacheLineIn;
					WriteEnable<=DWriteRequest;
				end
			end
			DELAY: begin
				Dstate<=VALID;
				Address<=DAddress;
				WriteData<=DWriteData;
				WriteEnable<=DWriteEnable;
				DValid<=1'b0;
			end
			VALID: begin
				Dstate<=NOREQUEST;
				DValid<=1'b1;
				WriteEnable<=1'b0;
			end
		endcase	

		
	end
end
	
	CacheMemories_1Word cms (
	.CLK(CLK),
	.Addr(Address),
	.DataIn(WriteData),
	.WE(WriteEnable),
	.DataOut(ReadData),
	.InvalidAddress(DCacheMemoryInvalidAddress)
	);
	
	assign InvalidAddress = DCacheMemoryInvalidAddress && (IValid || DValid); //assert invalid address if we're returning a result
// synopsys translate_on
endmodule

module CacheMemories_1Word (
	input CLK, 
	input [29:0] Addr,
	input [31:0] DataIn,
	output [31:0] DataOut,
	input WE,
	output InvalidAddress
) /* synthesis syn_black_box */ ;	
// synopsys translate_off
	
	localparam IROM_ADDR=22'b0,
	IRAM_ADDR =	17'b0000_0000_0100_0000_0,
	DRAM_ADDR = 17'b0001_0000_0000_0000_0,
	STACK_ADDR = 17'b0111_1111_1111_1111_1;
	
	reg [29:0] DelayAddr;
	
	
	wire IROM_SEL = (DelayAddr[29:8] == IROM_ADDR);
	wire IRAM_SEL = (DelayAddr[29:13] == IRAM_ADDR);
	wire DRAM_SEL = (DelayAddr[29:13] == DRAM_ADDR);
	wire STACK_SEL = (DelayAddr[29:13] == STACK_ADDR);
	
	wire IRAM0_SEL = IRAM_SEL && (DelayAddr[12:11] == 2'b00);
	wire IRAM1_SEL = IRAM_SEL && (DelayAddr[12:11] == 2'b01);
	wire IRAM2_SEL = IRAM_SEL && (DelayAddr[12:11] == 2'b10);
	wire IRAM3_SEL = IRAM_SEL && (DelayAddr[12:11] == 2'b11);
	
	wire DRAM0_SEL = DRAM_SEL && (DelayAddr[12:11] == 2'b00);
	wire DRAM1_SEL = DRAM_SEL && (DelayAddr[12:11] == 2'b01);
	wire DRAM2_SEL = DRAM_SEL && (DelayAddr[12:11] == 2'b10);
	wire DRAM3_SEL = DRAM_SEL && (DelayAddr[12:11] == 2'b11); 
	
	wire STACK0_SEL = STACK_SEL && (DelayAddr[12:11] == 2'b00);
	wire STACK1_SEL = STACK_SEL && (DelayAddr[12:11] == 2'b01);
	wire STACK2_SEL = STACK_SEL && (DelayAddr[12:11] == 2'b10);
	wire STACK3_SEL = STACK_SEL && (DelayAddr[12:11] == 2'b11);
	

	wire IROM_S = (Addr[29:8] == IROM_ADDR);
	wire IRAM_S = (Addr[29:13] == IRAM_ADDR);
	wire DRAM_S = (Addr[29:13] == DRAM_ADDR);
	wire STACK_S = (Addr[29:13] == STACK_ADDR);
	
	wire IRAM0_S = IRAM_S && (Addr[12:11] == 2'b00);
	wire IRAM1_S = IRAM_S && (Addr[12:11] == 2'b01);
	wire IRAM2_S = IRAM_S && (Addr[12:11] == 2'b10);
	wire IRAM3_S = IRAM_S && (Addr[12:11] == 2'b11);
	
	wire DRAM0_S = DRAM_S && (Addr[12:11] == 2'b00);
	wire DRAM1_S = DRAM_S && (Addr[12:11] == 2'b01);
	wire DRAM2_S = DRAM_S && (Addr[12:11] == 2'b10);
	wire DRAM3_S = DRAM_S && (Addr[12:11] == 2'b11); 
	
	wire STACK0_S = STACK_S && (Addr[12:11] == 2'b00);
	wire STACK1_S = STACK_S && (Addr[12:11] == 2'b01);
	wire STACK2_S = STACK_S && (Addr[12:11] == 2'b10);
	wire STACK3_S = STACK_S && (Addr[12:11] == 2'b11);

	wire IRAM0_WE = IRAM0_S && WE;
	wire IRAM1_WE = IRAM1_S && WE;
	wire IRAM2_WE = IRAM2_S && WE;
	wire IRAM3_WE = IRAM3_S && WE;
	
	wire DRAM0_WE = DRAM0_S && WE;
	wire DRAM1_WE = DRAM1_S && WE;
	wire DRAM2_WE = DRAM2_S && WE;
	wire DRAM3_WE = DRAM3_S && WE;
	
	wire STACK0_WE = STACK0_S && WE;
	wire STACK1_WE = STACK1_S && WE;
	wire STACK2_WE = STACK2_S && WE;
	wire STACK3_WE = STACK3_S && WE;
	
	wire [31:0] IRAM0_Data, IRAM1_Data, IRAM2_Data, IRAM3_Data;
	wire [31:0] DRAM0_Data, DRAM1_Data, DRAM2_Data, DRAM3_Data;
	wire [31:0] STACK0_Data, STACK1_Data, STACK2_Data, STACK3_Data;
	wire [31:0] IROM_Data;
	
	wire IA = ~(IRAM_SEL || DRAM_SEL || STACK_SEL || IROM_SEL);
	reg InvalidAddress;		 
	
	wire [31:0] IRAM_Data = IRAM0_SEL ? IRAM0_Data : (IRAM1_SEL ? IRAM1_Data : (IRAM2_SEL ? IRAM2_Data : IRAM3_Data));
	wire [31:0] DRAM_Data = DRAM0_SEL ? DRAM0_Data : (DRAM1_SEL ? DRAM1_Data : (DRAM2_SEL ? DRAM2_Data : DRAM3_Data));
	wire [31:0] STACK_Data = STACK0_SEL ? STACK0_Data : (STACK1_SEL ? STACK1_Data : (STACK2_SEL ? STACK2_Data : STACK3_Data));
	
	assign DataOut = IROM_SEL ? IROM_Data : (IRAM_SEL ? IRAM_Data : (DRAM_SEL ? DRAM_Data : (STACK_SEL ? STACK_Data : 31'b0)));
	
	
	//delay invalid address so the result is sent with the data output
	always @(posedge CLK) begin
		InvalidAddress<=IA;	   
		DelayAddr <= Addr;
	end
		
		
	CacheBIOSROM256x32 bios(
	.CLK(CLK),
	.ADDR(Addr[7:0]),
	.DataOut(IROM_Data)
	);
	
	Ramb2048x32 IRAM0 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(IRAM0_Data),
	.WE(IRAM0_WE)
	);			  
	
	Ramb2048x32 IRAM1 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(IRAM1_Data),
	.WE(IRAM1_WE)
	);			  

	Ramb2048x32 IRAM2 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(IRAM2_Data),
	.WE(IRAM2_WE)
	);			  

	Ramb2048x32 IRAM3 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(IRAM3_Data),
	.WE(IRAM3_WE)
	);			  

	
	Ramb2048x32 DRAM0 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(DRAM0_Data),
	.WE(DRAM0_WE)
	);
	
	Ramb2048x32 DRAM1 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(DRAM1_Data),
	.WE(DRAM1_WE)
	);
	
	Ramb2048x32 DRAM2 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(DRAM2_Data),
	.WE(DRAM2_WE)
	);
	
	Ramb2048x32 DRAM3 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(DRAM3_Data),
	.WE(DRAM3_WE)
	);
	
	Ramb2048x32 STACK0 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(STACK0_Data),
	.WE(STACK0_WE)
	);
					 
	Ramb2048x32 STACK1 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(STACK1_Data),
	.WE(STACK1_WE)
	);

	Ramb2048x32 STACK2 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(STACK2_Data),
	.WE(STACK2_WE)
	);

	Ramb2048x32 STACK3 (
	.CLK(CLK),
	.Addr(Addr[10:0]),
	.DataIn(DataIn),
	.DataOut(STACK3_Data),
	.WE(STACK3_WE)
	);	
// synopsys translate_on
endmodule

module CacheMemorySytem_4Word(
	input CLK, Reset, 
	input DCacheReadRequest, 
	input DCacheWriteRequest, 
	input [27:0] DCacheAddr,
	input [127:0] DCacheDataIn, 
	output [127:0] DCacheDataOut,
	output DCacheValid,
	output DCacheInvalidAddress,
	
	input ICacheReadRequest, 
	input [27:0] ICacheAddr,
	output [127:0] ICacheDataOut,
	output ICacheValid,
	output ICacheInvalidAddress,
	
	
	input BypassReadRequest,
	input [1:0] BypassWrite,
	input [31:0] BypassAddr,
	output [31:0] BypassDataOut,
	output BypassInvalidAddress,
	
	
	output [1:0] IODev0_WE, IODev1_WE, IODev2_WE, IODev3_WE, IOVGAChar_WE, IOVGAColor_WE,
	input [31:0] IODev0_DataIn, IODev1_DataIn, IODev2_DataIn, IODev3_DataIn
	
	) ;
	
	wire CacheInvalidAddress;
	assign DCacheInvalidAddress = CacheInvalidAddress;
	assign ICacheInvalidAddress = CacheInvalidAddress; 
	
	wire IO0_SEL,IO1_SEL,IO2_SEL,IO3_SEL,IOVGAChar_SEL,IOVGAColor_SEL;
	wire BypassWriteRequest = BypassWrite[1]|BypassWrite[0];
	
	IOMemory io (
	.Addr(BypassAddr),
	.Read(BypassReadRequest),
	.Write(BypassWriteRequest),
	.InvalidAddress(BypassInvalidAddress),
	.IODev0_SEL(IO0_SEL),
	.IODev1_SEL(IO1_SEL),
	.IODev2_SEL(IO2_SEL),
	.IODev3_SEL(IO3_SEL),
	.IOVGAChar_SEL(IOVGAChar_SEL),
	.IOVGAColor_SEL(IOVGAColor_SEL)
	);
	
	CacheMemoryInterface_4Word cache (
	.CLK(CLK),
	.Reset(Reset),
	.DReadRequest(DCacheReadRequest),
	.IReadRequest(ICacheReadRequest),
	.DWriteRequest(DCacheWriteRequest),
	.DAddr(DCacheAddr),
	.IAddr(ICacheAddr),			 
	.DCacheLineIn(DCacheDataIn),
	.DCacheLineOut(DCacheDataOut),
	.ICacheLineOut(ICacheDataOut),
	.DValid(DCacheValid),
	.IValid(ICacheValid),
	.InvalidAddress(CacheInvalidAddress)
	); 
	
	
	//logic below here to produce proper IO signals
	//--------------------------------------------------
	
	assign IODev0_WE = {IO0_SEL && BypassWrite[1], IO0_SEL && BypassWrite[0]}; 
	assign IODev1_WE = {IO1_SEL && BypassWrite[1], IO1_SEL && BypassWrite[0]}; 
	assign IODev2_WE = {IO2_SEL && BypassWrite[1], IO2_SEL && BypassWrite[0]}; 
	assign IODev3_WE = {IO3_SEL && BypassWrite[1], IO3_SEL && BypassWrite[0]}; 
	
	assign IOVGAChar_WE = {IOVGAChar_SEL && BypassWrite[1], IOVGAChar_SEL && BypassWrite[0]};
	assign IOVGAColor_WE = {IOVGAColor_SEL && BypassWrite[1], IOVGAColor_SEL && BypassWrite[0]};
	
	assign BypassDataOut = IO0_SEL ? IODev0_DataIn : (IO1_SEL ? IODev1_DataIn : (IO2_SEL ? IODev2_DataIn : IODev3_DataIn));
	
	
endmodule

/*Assert Read/Write Request high for one clock cycle and then read result when valid line goes high.*/
module CacheMemoryInterface_4Word(input CLK, Reset, 
	input DReadRequest, DWriteRequest,
	input [27:0] DAddr, input [127:0] DCacheLineIn, output [127:0] DCacheLineOut, output DValid, output InvalidAddress,
	input IReadRequest,
	input [27:0] IAddr, output [127:0] ICacheLineOut, output IValid
	)  ;
	
	
	
	
	localparam NOREQUEST=1,
	DELAY=2,
	VALID=4;
	
	reg [4:0] Istate;
	reg [4:0] Dstate;
	
	reg IValid, DValid;	
	reg WriteEnable;
	reg [27:0] Address;			 
	reg [27:0] IAddress;
	reg [27:0] DAddress;
	
	reg [127:0] WriteData;
	reg [127:0] DWriteData;
	reg	DWriteEnable; 
	
	wire [127:0] ReadData;
	
	wire DCacheMemoryInvalidAddress;
	
	assign DCacheLineOut = ReadData;
	assign ICacheLineOut = ReadData;
	
	
	always @(posedge CLK or posedge Reset) begin
		if (Reset) begin
			Istate<=NOREQUEST;
			Dstate<=NOREQUEST;
			IAddress<=28'b0;
			DAddress<=28'b0;
			IValid<=1'b0;
			DValid<=1'b0;
			WriteEnable<= 1'b0;
			WriteData<=128'b0;
			DWriteData<=128'b0;
			DWriteEnable<=1'b0;
			Address<=28'b0;
		end	else begin
		
		case (Istate)
			NOREQUEST: begin
				IValid<=1'b0;
				if (Dstate==DELAY && IReadRequest) begin
					//there's a data cache request that was delayed for an instruction cache request, so let that finish first
					Istate<=DELAY;
					IAddress<=IAddr;
				end else if (Dstate!=DELAY && IReadRequest) begin
					//no data request in progress, go ahead and request directly
					Istate<=VALID;
					Address<=IAddr;
					WriteEnable<=1'b0;
				end
			end
			DELAY: begin			//we wanted to wait a cycle before requesting the data, so go ahead and request now
				Istate<=VALID;
				Address<=IAddress;
				WriteEnable<=1'b0;
				IValid<=1'b0;
			end				  	  
			VALID: begin			//read request completed
				Istate<=NOREQUEST;
				IValid<=1'b1;
			end
		endcase
		
		case (Dstate)
			NOREQUEST: begin	
				DValid<=1'b0;
				if (Istate==DELAY && (DReadRequest || DWriteRequest)) begin
					//there a delayed instruction request, finish that first, delay the data request
					Dstate<=DELAY;
					DAddress<=DAddr;
					DWriteData<=DCacheLineIn;
					DWriteEnable<=DWriteRequest;
				end else if (Istate!=DELAY && (DReadRequest || DWriteRequest) && IReadRequest) begin 
					//no instructions queued, but we give priority to the instruction cache, so delay the instruction request
					Dstate<=DELAY;
					DAddress<=DAddr;
					DWriteData<=DCacheLineIn;
					DWriteEnable<=DWriteRequest;
				end else if (Istate!=DELAY && (DReadRequest || DWriteRequest) && ~IReadRequest) begin 
					//data request can execute immediately
					Dstate<=VALID;
					Address<=DAddr;
					WriteData<=DCacheLineIn;
					WriteEnable<=DWriteRequest;
				end
			end
			DELAY: begin
				Dstate<=VALID;
				Address<=DAddress;
				WriteData<=DWriteData;
				WriteEnable<=DWriteEnable;
				DValid<=1'b0;
			end
			VALID: begin
				Dstate<=NOREQUEST;
				DValid<=1'b1;
				WriteEnable<=1'b0;
			end
		endcase	

		
	end
end
	
	CacheMemories_4Word cms (
	.CLK(CLK),
	.Addr(Address),
	.DataIn(WriteData),
	.WE(WriteEnable),
	.DataOut(ReadData),
	.InvalidAddress(DCacheMemoryInvalidAddress)
	);
	
	assign InvalidAddress = DCacheMemoryInvalidAddress && (IValid || DValid); //assert invalid address if we're returning a result
	
endmodule

module CacheMemories_4Word (input CLK, input [27:0] Addr, input [127:0] DataIn, output [127:0] DataOut, input WE, output InvalidAddress) ; 
	
	localparam IROM_ADDR=22'b0,
	IRAM_ADDR =	17'b0000_0000_0100_0000_0,
	DRAM_ADDR = 17'b0001_0000_0000_0000_0,
	STACK_ADDR = 17'b0111_1111_1111_1111_1;
	
	reg [27:0] DelayAddr;
	
	
	wire IROM_SEL = (DelayAddr[27:6] == IROM_ADDR);
	wire IRAM_SEL = (DelayAddr[27:11] == IRAM_ADDR);
	wire DRAM_SEL = (DelayAddr[27:11] == DRAM_ADDR);
	wire STACK_SEL = (DelayAddr[27:11] == STACK_ADDR);
	
	wire IRAM0_SEL = IRAM_SEL && (DelayAddr[10:9] == 2'b00);
	wire IRAM1_SEL = IRAM_SEL && (DelayAddr[10:9] == 2'b01);
	wire IRAM2_SEL = IRAM_SEL && (DelayAddr[10:9] == 2'b10);
	wire IRAM3_SEL = IRAM_SEL && (DelayAddr[10:9] == 2'b11);
	
	wire DRAM0_SEL = DRAM_SEL && (DelayAddr[10:9] == 2'b00);
	wire DRAM1_SEL = DRAM_SEL && (DelayAddr[10:9] == 2'b01);
	wire DRAM2_SEL = DRAM_SEL && (DelayAddr[10:9] == 2'b10);
	wire DRAM3_SEL = DRAM_SEL && (DelayAddr[10:9] == 2'b11); 
	
	wire STACK0_SEL = STACK_SEL && (DelayAddr[10:9] == 2'b00);
	wire STACK1_SEL = STACK_SEL && (DelayAddr[10:9] == 2'b01);
	wire STACK2_SEL = STACK_SEL && (DelayAddr[10:9] == 2'b10);
	wire STACK3_SEL = STACK_SEL && (DelayAddr[10:9] == 2'b11);
	

	wire IROM_S = (Addr[27:6] == IROM_ADDR);
	wire IRAM_S = (Addr[27:11] == IRAM_ADDR);
	wire DRAM_S = (Addr[27:11] == DRAM_ADDR);
	wire STACK_S = (Addr[27:11] == STACK_ADDR);
	
	wire IRAM0_S = IRAM_S && (Addr[10:9] == 2'b00);
	wire IRAM1_S = IRAM_S && (Addr[10:9] == 2'b01);
	wire IRAM2_S = IRAM_S && (Addr[10:9] == 2'b10);
	wire IRAM3_S = IRAM_S && (Addr[10:9] == 2'b11);
	
	wire DRAM0_S = DRAM_S && (Addr[10:9] == 2'b00);
	wire DRAM1_S = DRAM_S && (Addr[10:9] == 2'b01);
	wire DRAM2_S = DRAM_S && (Addr[10:9] == 2'b10);
	wire DRAM3_S = DRAM_S && (Addr[10:9] == 2'b11); 
	
	wire STACK0_S = STACK_S && (Addr[10:9] == 2'b00);
	wire STACK1_S = STACK_S && (Addr[10:9] == 2'b01);
	wire STACK2_S = STACK_S && (Addr[10:9] == 2'b10);
	wire STACK3_S = STACK_S && (Addr[10:9] == 2'b11);

	wire IRAM0_WE = IRAM0_S && WE;
	wire IRAM1_WE = IRAM1_S && WE;
	wire IRAM2_WE = IRAM2_S && WE;
	wire IRAM3_WE = IRAM3_S && WE;
	
	wire DRAM0_WE = DRAM0_S && WE;
	wire DRAM1_WE = DRAM1_S && WE;
	wire DRAM2_WE = DRAM2_S && WE;
	wire DRAM3_WE = DRAM3_S && WE;
	
	wire STACK0_WE = STACK0_S && WE;
	wire STACK1_WE = STACK1_S && WE;
	wire STACK2_WE = STACK2_S && WE;
	wire STACK3_WE = STACK3_S && WE;
	
	wire [127:0] IRAM0_Data, IRAM1_Data, IRAM2_Data, IRAM3_Data;
	wire [127:0] DRAM0_Data, DRAM1_Data, DRAM2_Data, DRAM3_Data;
	wire [127:0] STACK0_Data, STACK1_Data, STACK2_Data, STACK3_Data;
	wire [127:0] IROM_Data;
	
	wire IA = ~(IRAM_SEL || DRAM_SEL || STACK_SEL || IROM_SEL);
	reg InvalidAddress;		 
	
	wire [127:0] IRAM_Data = IRAM0_SEL ? IRAM0_Data : (IRAM1_SEL ? IRAM1_Data : (IRAM2_SEL ? IRAM2_Data : IRAM3_Data));
	wire [127:0] DRAM_Data = DRAM0_SEL ? DRAM0_Data : (DRAM1_SEL ? DRAM1_Data : (DRAM2_SEL ? DRAM2_Data : DRAM3_Data));
	wire [127:0] STACK_Data = STACK0_SEL ? STACK0_Data : (STACK1_SEL ? STACK1_Data : (STACK2_SEL ? STACK2_Data : STACK3_Data));
	
	assign DataOut = IROM_SEL ? IROM_Data : (IRAM_SEL ? IRAM_Data : (DRAM_SEL ? DRAM_Data : (STACK_SEL ? STACK_Data : 128'b0)));
	
	
	//delay invalid address so the result is sent with the data output
	always @(posedge CLK) begin
		InvalidAddress<=IA;	   
		DelayAddr <= Addr;
	end
		
		
	CacheBIOSROM64x128 bios(
	.CLK(CLK),
	.ADDR(Addr[5:0]),
	.DataOut(IROM_Data)
	);
	
	Ramb512x128 IRAM0 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(IRAM0_Data),
	.WE(IRAM0_WE)
	);			  
	
	Ramb512x128 IRAM1 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(IRAM1_Data),
	.WE(IRAM1_WE)
	);			  

	Ramb512x128 IRAM2 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(IRAM2_Data),
	.WE(IRAM2_WE)
	);			  

	Ramb512x128 IRAM3 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(IRAM3_Data),
	.WE(IRAM3_WE)
	);			  

	
	Ramb512x128 DRAM0 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(DRAM0_Data),
	.WE(DRAM0_WE)
	);
	
	Ramb512x128 DRAM1 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(DRAM1_Data),
	.WE(DRAM1_WE)
	);
	
	Ramb512x128 DRAM2 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(DRAM2_Data),
	.WE(DRAM2_WE)
	);
	
	Ramb512x128 DRAM3 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(DRAM3_Data),
	.WE(DRAM3_WE)
	);
	
	Ramb512x128 STACK0 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(STACK0_Data),
	.WE(STACK0_WE)
	);
					 
	Ramb512x128 STACK1 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(STACK1_Data),
	.WE(STACK1_WE)
	);

	Ramb512x128 STACK2 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(STACK2_Data),
	.WE(STACK2_WE)
	);

	Ramb512x128 STACK3 (
	.CLK(CLK),
	.Addr(Addr[8:0]),
	.DataIn(DataIn),
	.DataOut(STACK3_Data),
	.WE(STACK3_WE)
	);	
	
endmodule



